#!/usr/bin/perl -I/usr/sausalito/perl
#
# $Id: vz_vserver_backup.pl
#
# This script performs local and remote backing up of virtual private servers. 
# Only VPS's which have either local or remote backup enabled in the GUI will be backed up!
#

use CCE;
use POSIX qw(isalpha);
use Socket;
use Sys::Hostname;
use File::Spec;
use List::MoreUtils qw(uniq);
use POSIX qw(strftime);
use MIME::Lite;
use File::Basename;

my $cce = new CCE;
$cce->connectuds();

$oid = $cce->event_oid();
$obj = $cce->event_object();

# Helper to redirect SSH errors to /dev/null:
open $ssherr, '>', '/dev/null'; # or die "Unable to open /dev/null";

$host = hostname();
$addr = inet_ntoa(scalar(gethostbyname($host)) || 'localhost');
$now = localtime time;
$datetime = strftime('%Y-%m-%d %H:%M:%S', localtime);

$DEBUG = "1";

# Get local / remote switch:
$type_switch = $ARGV[0];
if ($type_switch eq "local") {
    $b_type = "local";
    $B_type = "Local";
    $s_type = "local";
}
elsif ($type_switch eq "remote") {
    $b_type = "remote";
    $B_type = "Remote";
    $s_type = "remote";
}
else {
    $b_type = "local & remote";
    $B_type = "Local & Remote";
    $s_type = "both";
}

open(CONF, "< /etc/swatch.conf");
@email_list;
while (<CONF>) {
  chomp;
  my($key, $val) = split /\s*=\s*/, $_, 2;
  if ($key eq "email_list") {
    @email_list = split /\s*,\s*/, $val;
  } elsif ($key eq "lang") {
    $lang = $val;
  } elsif ($key eq "enabled") {
    $enabled = $val;
  }
}

# Find out if there is anything to back up (we may be cluster secondary after all!)
if (!-d "/vz/private") {
    print STDERR "INFO: This node is currently secondary in a cluster.\n";
    print STDERR "So there is nothing to backup on this node.\n\n";
    exit 1;
}

# Get settings out of CODB to determine which type of backup is enabled:
@oids = $cce->find('VserverMasterNode');
if (!defined($oids[0])) {
    print STDERR "VZBACKUP: Sorry, no VserverMasterNode object in CCE found!\n";
    exit 1;
}

# Remove stale lockfiles if they're older than 180 minutes:  
if (-f "/tmp/vz_vserver_backup.lock") {
    system("find /tmp/vz_vserver_backup.lock -type f -cmin +180 -print | xargs rm >/dev/null 2>&1");
}

# Create lockfile:
if (-f "/tmp/vz_vserver_backup.lock") {
    print STDERR "ERROR: Found lockfile /tmp/vz_vserver_backup.lock!\n";
    print STDERR "Please check that no other instance of this program runs!\n\n";
    exit 1;
}

    # Email generation:
    open (EMAIL, '>>/tmp/.backup_mail');
    print EMAIL "=================================================================\n";
    print EMAIL "Aventurin{e} $B_type Backup on $host ($addr) \n";
    print EMAIL "Starting up at: $now\n";
    print EMAIL "=================================================================\n\n";


($ok, $solbackupValues) = $cce->get($oids[0]);
$type = 'none';
if ($ok) {
    if (($solbackupValues->{'MN_LocalBackupEnabled'} eq "1") && ($s_type eq 'local') || ($s_type eq 'both')) {
        $type = 'local';
        &do_local_backup;
    }
    if (($solbackupValues->{'MN_RemoteBackupEnabled'} eq "1") && ($s_type eq 'remote') || ($s_type eq 'both')) {
        $type = 'remote';
        &do_remote_backup;
    }
    if ($type eq 'none') {
        # Do nothing:
        print STDERR "INFO: Backup is not configured on this node.\n";
        print STDERR "So there is nothing to do here. Please configure\n";
        print STDERR "the backup through the GUI first!\n";
    }
    # Remove lockfile:
    system("/bin/rm -f /tmp/vz_vserver_backup.lock");
}

if (-f '/tmp/.backup_mail') {

    $now = localtime time;
    print EMAIL "=================================================================\n";
    print EMAIL "Aventurin{e} $B_type Backup on $host ($addr) \n";
    print EMAIL "Backup finished at: $now\n";
    print EMAIL "=================================================================\n\n";
    close (EMAIL);

    # Send the email:
    open (EMAIL, '/tmp/.backup_mail');
    while (<EMAIL>) {
        $body .= $_;
    }
    close (EMAIL);
    system("/bin/rm -f /tmp/.backup_mail");

    if ($body) {
        $body = $body_head . $body;
        $subject = $host . ": VPS backup finished";

        my $to;
        foreach $to (@email_list) {

            # Build the message using MIME::Lite instead:
            my $send_msg = MIME::Lite->new(
                    From     => "root@$host",
                    To       => "$to",
                    Subject  => $subject,
                    Data     => $body
            );

            # Set content type:
            $send_msg->attr("content-type"         => "text/plain");
            $send_msg->attr("content-type.charset" => "ISO-8859-1"); 
             
            # Out with the email:
            $send_msg->send;
        }
    }
}

$cce->bye('SUCCESS');
exit 0;

#
### Subroutines:
#

# Local backup:
sub do_local_backup {

    $RetainDays_local = $solbackupValues->{'RetainDays_local'};

    $localLocation = $solbackupValues->{'MN_BackupLocalLocation'} . '/local/';
    if (!-d $localLocation) {
        system("mkdir -p $localLocation");
    }

    $workDir = $solbackupValues->{'MN_BackupLocalLocation'} . '/tmp/';
    if (!-d $workDir) {
        system("mkdir -p $workDir");
    }

    # Get list of VPS's with enabled backup:
    @LocalBackups = $cce->findx('Vserver_Backup', {'virttype' => 'openvz', 'VserverBackupLocalEnabled' => '1'}, {}, '', '');

    print STDERR "Starting local backup of VPS's ... \n\n";
    print EMAIL "Starting local backup of VPS's ... \n";
    print EMAIL "=========================================\n\n";

    $localNum = '0';
    foreach my $vpsOID (@LocalBackups) {
        (my $ok, $vps) = $cce->get($vpsOID);

        $vpsid = $vps->{'vpsid'};
        $vpsName = `/usr/bin/prlctl list $vpsid -H -o hostname`;
        chomp($vpsName);

        print STDERR "Performing local backup of VPS $vpsid ('" . $vpsName . "').\n";
        print EMAIL "Performing local backup of VPS $vpsid ('" . $vpsName . "').\n";

        if (($localLocation ne '') || ($localLocation ne '/')) {
            if ($RetainDays_local == '1') {
                $fileNameToDelete = $localLocation . '/' . "vzpbackup*_" . $vpsid . "_*.tar";
                print STDERR "Deleting: $fileNameToDelete\n";
                system("rm -f $fileNameToDelete");
            }
            else {
                if (($localLocation ne '') || ($localLocation ne '/')) {
                    #print STDERR "Running: find $localLocation/ -type f -mtime +$RetainDays_local -name 'vzpbackup*.tar' -execdir rm -- '{}' +\n";
                    system("find $localLocation/ -type f -mtime +$RetainDays_local -name 'vzpbackup*.tar' -execdir rm -- '{}' +");
                }
            }
        }

        # vzpbackup.sh [--suspend=<yes/no>] [--backup-dir=<Backup-Directory>] [--work-dir=<Temp-Directory>] [--compress=<no/pz/bz/pbz/tbz/gz/tgz/xz/txz>] [--compact] [--all] <CTID> <CTID>"
        system("/usr/sausalito/sbin/vzpbackup.sh --suspend=no --backup-dir=$localLocation --work-dir=$workDir --compress=no $vpsid >/dev/null 2>&1 || :");

        $remDir = $localLocation . $vpsid;
        if (-d $remDir) {
            system("rm -Rf $remDir");
        }

        print STDERR "Finished local backup of VPS $vpsid ('" . $vpsName . "').\n\n";
        print EMAIL "Finished local backup of VPS $vpsid ('" . $vpsName . "').\n\n";

        $localNum++;
    }
    if ($localNum == '0') {
        print STDERR "No VPS has local backup enabled. \n";
        print EMAIL "No VPS has local backup enabled. \n\n";        
    }
}

# Remote backup:
sub do_remote_backup {

    $RetainDays_remote = $solbackupValues->{'RetainDays_remote'};

    $remoteLocation = $solbackupValues->{'MN_BackupLocalLocation'} . '/remote';
    if (!-d $remoteLocation) {
        system("mkdir -p $remoteLocation");
    }

    $workDir = $solbackupValues->{'MN_BackupLocalLocation'} . '/tmp/';
    if (!-d $workDir) {
        system("mkdir -p $workDir");
    }

    $rhost = $solbackupValues->{'MN_BackupRemoteTargetHost'};
    $MNBport = $solbackupValues->{'MN_BackupRemoteTargetPort'};
    if ($MNBport eq '22') {
        $addr = 'root@' . $rhost . ':' . $solbackupValues->{'MN_BackupRemoteLocation'};
    }
    else {
        $addr = 'root@' . $rhost . ':' . $solbackupValues->{'MN_BackupRemoteLocation'} . ' -p ' . $MNBport;
    }

    $checkMount = `/usr/bin/mount | /usr/bin/grep $rhost | /usr/bin/grep $remoteLocation | /usr/bin/wc -l`;
    chomp($checkMount);

    if ($checkMount eq '0') {
        # Not mounted!
        system("/usr/sbin/modprobe fuse");
        system("/usr/bin/sshfs $addr $remoteLocation");
    }

    $checkMount = `/usr/bin/mount | /usr/bin/grep $rhost | /usr/bin/grep $remoteLocation | /usr/bin/wc -l`;
    chomp($checkMount);

    if ($checkMount eq '0') {
        print STDERR "ERROR: Unable to mount $addr at $remoteLocation! \n\n";
        print EMAIL "ERROR: Unable to mount $addr at $remoteLocation! \n";
    }

    # Get list of VPS's with enabled backup:
    @LocalBackups = $cce->findx('Vserver_Backup', {'virttype' => 'openvz', 'VserverBackupRemoteEnabled' => '1'}, {}, '', '');

    # Email generation:
    print STDERR "Starting remote backup of VPS's ... \n\n";
    print EMAIL "Starting remote backup of VPS's ... \n";
    print EMAIL "=========================================\n\n";

    $remoteNum = '0';
    foreach my $vpsOID (@LocalBackups) {
        (my $ok, $vps) = $cce->get($vpsOID);

        $vpsid = $vps->{'vpsid'};
        $vpsName = `/usr/bin/prlctl list $vpsid -H -o hostname`;
        chomp($vpsName);

        print STDERR "Performing remote backup of VPS $vpsid ('" . $vpsName . "').\n";
        print EMAIL "Performing remote backup of VPS $vpsid ('" . $vpsName . "').\n";

        #print STDERR "RetainDays_remote: $RetainDays_remote\n";
        #print STDERR "remoteLocation: $remoteLocation\n";

        if (($remoteLocation ne '') || ($remoteLocation ne '/')) {
            if ($RetainDays_remote == '1') {
                $fileNameToDelete = $remoteLocation . '/' . "vzpbackup*_" . $vpsid . "_*.tar";
                print STDERR "Deleting: $fileNameToDelete\n";
                system("rm -f $fileNameToDelete");
            }
            else {
                if (($remoteLocation ne '') || ($remoteLocation ne '/')) {
                    #print STDERR "Running: find $remoteLocation/ -type f -mtime +$RetainDays_remote -name 'vzpbackup*.tar' -execdir rm -- '{}' +\n";
                    system("find $remoteLocation/ -type f -mtime +$RetainDays_remote -name 'vzpbackup*.tar' -execdir rm -- '{}' +");
                }
            }
        }

        # vzpbackup.sh [--suspend=<yes/no>] [--backup-dir=<Backup-Directory>] [--work-dir=<Temp-Directory>] [--compress=<no/pz/bz/pbz/tbz/gz/tgz/xz/txz>] [--compact] [--all] <CTID> <CTID>"
        system("/usr/sausalito/sbin/vzpbackup.sh --suspend=no --backup-dir=$remoteLocation --work-dir=$workDir --compress=no $vpsid >/dev/null 2>&1 || :");

        $remDir = $remoteLocation . '/' . $vpsid;
        if (-d $remDir) {
            system("rm -Rf $remDir");
        }

        print STDERR "Finished remote backup of VPS $vpsid ('" . $vpsName . "').\n\n";
        print EMAIL "Finished remote backup of VPS $vpsid ('" . $vpsName . "').\n\n";

        $remoteNum++;
    }
    if ($remoteNum == '0') {
        print STDERR "No VPS has remote backup enabled. \n";
        print EMAIL "No VPS has remote backup enabled. \n\n";        
    }
}

# 
# Copyright (c) 2015-2020 Michael Stauber, SOLARSPEED.NET
# All Rights Reserved.
# 
# 1. Redistributions of source code must retain the above copyright 
#     notice, this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright 
#     notice, this list of conditions and the following disclaimer in 
#     the documentation and/or other materials provided with the 
#     distribution.
# 
# 3. Neither the name of the copyright holder nor the names of its 
#     contributors may be used to endorse or promote products derived 
#     from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
# 
# You acknowledge that this software is not designed or intended for 
# use in the design, construction, operation or maintenance of any 
# nuclear facility.
# 
